/* ======================================================================
     "libbx1CP2.cc"
       - basis expansion library for cos(theta) power series 2-D
          potential and associated Legendre basis
             Copyright (c) 2004-2006  by A. Miyoshi, Univ. Tokyo
                         created: June  19, 2006
                     last edited: July  19, 2006   Rev. 0.01
 ====================================================================== */

#include "libbx1CP2.h"

/* ======================================================================
     power series potential
 ====================================================================== */

int powerSeries::maxOrder() {                        // --- maxOrder ---
  int mxord = 0;
  iterator p;
  for (p = begin(); p != end(); p++)
   { if (p->first > mxord) { mxord = p->first; } }
  return mxord;
}

powerSeries powerSeries::operator/(double den) {          // --- "/" ---
  powerSeries res = *this;
  iterator p;
  for (p = res.begin(); p != res.end(); p++) { p->second /= den; }
  return res;
}

vector<double> powerSeries::coefVect() {             // --- coefVect ---
  vector<double> res(maxOrder() + 1, 0.);
  iterator p;
  for (p = begin(); p != end(); p++) { res[p->first] = p->second; }
  return res;
}

/* ======================================================================
     associated Legendre basis and related classes
 ====================================================================== */

/* ----- global functions ------------------------------------------- */

  // --- factorial of an interger
double intFactorial(int x) {                     // --- intFactorial ---
  int i;
  double rs = 1.;
  if (x < 2) { return rs; }
  for (i = x; i > 1; i--) { rs *= i; }
  return rs;
}

  // --- double factorial of an integer
double intDblFactrl(int x) {                     // --- intDblFactrl ---
  int i;
  double rs = 1.;
  if (x < 2) { return rs; }
  for (i = x; i > 1; i -= 2) { rs *= i; }
  return rs;
}

  // --- Knm coefficient for NnmPnm
double recK(int n, int m) {                              // --- recK ---
  if ((n < 1) || (m < 0) || (m > n)) { return 0.; }
  return sqrt( (double)(n + m) / (2. * n + 1)
             * (double)(n - m) / (2. * n - 1) );
}

  // --- normalized Pnn(x)
double nrmPnn(int n, double x) {                       // --- nrmPnn ---
  int sign = 1;
  double rs = 0.;
  if ((n < 0) || (x * x > 1.)) { return rs; }
  if ((n % 2) == 1) { sign = -1; }
  rs = sign * intDblFactrl(2 * n + 1) / sqrt(2 * intFactorial(2 * n + 1));
  rs *= pow(1. - x * x, n / 2.);
  return rs;
}

  // --- normalized Pnm(x)
vector<double> nrmPnm(int m, int nmax, double x) {     // --- nrmPnm ---
  int n;
  double p, p1, p2;
  vector<double> rs;
  rs.clear();
   if ((nmax < m) || (m < 0)) { return rs; }
  p2 = nrmPnn(m, x); rs.push_back(p2);
   if (nmax == m) { return rs; }
  p1 = x * p2 / recK(m + 1, m); rs.push_back(p1);
   if (nmax == (m + 1)) { return rs; }
  for (n = m + 2; n <= nmax; n++) {
    p = (x * p1 - recK(n - 1, m) * p2) / recK(n, m);
    rs.push_back(p); p2 = p1; p1 = p;
  }
  return rs;
}

/* ----- recurrent power expansion coefficients --------------------- */

                                                 // --- initializers ---
recurExpnCoef::recurExpnCoef() { clear(); }

recurExpnCoef::recurExpnCoef(int m, int jmx, int kc) { init(m, jmx, kc); }

void recurExpnCoef::clear() {
  vector<double>::clear();
  mOrd = 0; jMax = 0; kCent = 0; iPow = 0;
}

void recurExpnCoef::init(int m, int jmx, int kc) {
  int siz;
  clear();
  mOrd = m; jMax = jmx; kCent = kc; iPow = 0;
  resize(jMax - mOrd + 1, 0.);
  (*this)[kCent - mOrd] = 1.;
}

void recurExpnCoef::incPow() {                         // --- incPow ---
  int n, p;
  vector<double> cnx(jMax - mOrd + 1, 0.);
  for (n = kCent - iPow; n <= kCent + iPow; n += 2) {
    if ((n >= mOrd) && (n <= jMax)) {
      if ((n - 1) >= mOrd)
       { cnx[n - 1 - mOrd] += (*this)[n - mOrd] * recK(n, mOrd); }
      if ((n + 1) <= jMax)
       { cnx[n + 1 - mOrd] += (*this)[n - mOrd] * recK(n + 1, mOrd); }
    }
  }
  for (p = 0; p <= jMax - mOrd; p++) { (*this)[p] = cnx[p]; }
  iPow++;
}
                                                        // --- addTo ---
void recurExpnCoef::addTo(vector<double> &c, double f) {
  int p, siz = size();
  if (c.size() < siz) { siz = c.size(); }
  for (p = 0; p < siz; p++) { c[p] += f * (*this)[p]; }
}

/* ----- eigenvalues for 2-D symmetric rotor ------------------------ */

void eigenVal2Dsrot::clear() {                    // --- initializer ---
  vector<vector<double> >::clear();
  qnTable.clear();
}

int eigenVal2Dsrot::numEgnv() {                       // --- numEgnv ---
  return qnTable.size();
}

void eigenVal2Dsrot::evalSort() {                    // --- evalSort ---
  int nEgnv = qnTable.size(), i, j;
  double vi, vj;
  for (i = 0; i < (nEgnv - 1); i++) {
    vi = getEgnv(i);
    for (j = i + 1; j < nEgnv; j++) {
      vj = getEgnv(j);
      if (vj < vi)
       { std::swap(qnTable[i], qnTable[j]); std::swap(vi, vj); }
    }
  }
}

double eigenVal2Dsrot::getEgnv(int idx) {             // --- getEgnv ---
  int m, j;
  if (idx >= qnTable.size()) { return 0.; }
  m = qnTable[idx].first; j = qnTable[idx].second;
  return (*this)[m][j];
}

/* ----- associated Legendre basis ---------------------------------- */

                                                 // --- initializers ---
basisAssocLegendre::basisAssocLegendre() { clear(); }
basisAssocLegendre::basisAssocLegendre(int jmx, int ojm, double rtc,
 nucSpinStat &nsw, powerSeries &pe) { init(jmx, ojm, rtc, nsw, pe); }

void basisAssocLegendre::clear() {
  maxJdeg = 0; outJmax = 0; rotB = 0.;
  nSpinW.clear(); potE.clear(); mJcur = -1;
}

void basisAssocLegendre::init(int jmx, int ojm, double rtc,
 nucSpinStat &nsw, powerSeries &pe) {
  maxJdeg = jmx; outJmax = ojm; rotB = rtc; nSpinW = nsw;
  potE = pe / rotB; mJcur = -1;
}

int basisAssocLegendre::solveOnly() {               // --- solveOnly ---
  int m, nEgnv, ie;
  vector<double> eg;
  egnVal.clear();
  for (m = 0; m <= maxJdeg; m++) {
    if (solveForOneM(m) == ERRE) { return ERRE; }
    nEgnv = maxJdeg - m + 1; eg.clear();
    for (ie = 0; ie < nEgnv; ie++) {
      eg.push_back(w[ie]); egnVal.qnTable.push_back(make_pair(m, ie));
    }
    egnVal.push_back(eg);
  }
  egnVal.evalSort();
  return NORM;
}
                                                   // --- solveWrite ---
int basisAssocLegendre::solveWrite(string &basfln, int outCtrl) {
  int m, nEgnv, ie;
  vector<double> eg;
  egnVal.clear();
  for (m = 0; m <= maxJdeg; m++) {
    if (solveForOneM(m) == ERRE) { return ERRE; }
    nEgnv = maxJdeg - m + 1; eg.clear();
    for (ie = 0; ie < nEgnv; ie++) {
      eg.push_back(w[ie]); egnVal.qnTable.push_back(make_pair(m, ie));
    }
    egnVal.push_back(eg);
    if (m <= outJmax) {
      if (outCtrl & OUTvec)
       { if (writeVectors(basfln) == ERRE) { return ERRE; } }
      if (outCtrl & OUTfunc)
       { if (writeFunctions(basfln) == ERRE) { return ERRE; } }
    }
  }
  egnVal.evalSort();
  return NORM;
}

                                                // --- writeEnergies ---
int basisAssocLegendre::writeEnergies(string &basfln, int maxRNE) {
  int symn = nSpinW.size(), iss, sumSw = 0;
  int nEgnv = egnVal.qnTable.size(), ie, mJ, ij, J;
  string outfln = basfln + "_eigen_values.csv";
  ofstream ouf;
  outFormat *ofmt;
  for (iss = 0; iss < symn; iss++) { sumSw += nSpinW[iss]; }
  ouf.open(outfln.c_str());
  if (!ouf) { return ERRE; }
  ofmt = new outFormat(&ouf);
  ouf << "number,eigenValue,J,mJ,w\n";
  ouf.precision(8);
  for (ie = 0; ie < nEgnv; ie++) {
    mJ = egnVal.qnTable[ie].first;
    ij = egnVal.qnTable[ie].second;
    J = mJ + ij; iss = J % symn;
    ouf << ie << ',' << egnVal[mJ][ij] * rotB << ',' << J << ',' << mJ
     << ",(" << nSpinW[iss] << '/' << sumSw << ")";
    if (ie == maxRNE) { ouf << ",(maximum reliable eigenvalue)"; }
    ouf << endl;
  }
  delete ofmt;
  ouf.close();
  return NORM;
}

                                               // --- writePartFuncs ---
int basisAssocLegendre::writePartFuncs(string &basfln, int maxRNE,
 tempList &tlist) {
  int symn = nSpinW.size(), iss, sumSw = 0;
  int ieMax, itl, ntl = tlist.size(), ie, mJ, ij, J;
  double egvMax, Tmax, TmaxInList, T, kT, errPf, Qpf, zpe, QpfZ, mDeg;
  ofstream ouf;
  string outfln = basfln + "_part_funcs.csv";
  outFormat *ofmf;

  for (iss = 0; iss < symn; iss++) { sumSw += nSpinW[iss]; }
  zpe = egnVal.getEgnv(0) * rotB;
  if (maxRNE < 0) {
    ieMax = egnVal.numEgnv() - 1;
    Tmax = tlist.maxTemp();
  } else {
    egvMax = egnVal.getEgnv(maxRNE) * rotB;
    Tmax = egvMax * KELWAVN / log(100.);
    if ((TmaxInList = tlist.maxLowerEq(Tmax)) < 0.)
     { return errms(LRGQerr); }
    errPf = exp(-egvMax * KELWAVN / TmaxInList);
    cout << "estimated partition function error is " << errPf << " at "
     << TmaxInList << " K.\n";
    if (TmaxInList < tlist.maxTemp()) { errms(LMTQrng); }
    ieMax = maxRNE;
  }
  ouf.open(outfln.c_str());
  if (!ouf) { return errms(FAILopen, outfln); }
  ofmf = new outFormat(&ouf);
  ouf.precision(8);
  ouf << "10000/T,T,Q,Qz\n";
  for (itl = 0; itl < ntl; itl++) {
    T = tlist[itl];
    if (T <= Tmax) {
      kT = T / KELWAVN; Qpf = 0.;
      for (ie = ieMax; ie >= 0; ie--) {
        mJ = egnVal.qnTable[ie].first;
        ij = egnVal.qnTable[ie].second;
        J = mJ + ij; iss = J % symn;
        mDeg = 1.;
        if (mJ > 0) { mDeg = 2.; }
        Qpf += (double)(nSpinW[iss]) * mDeg
               * exp(-egnVal[mJ][ij] * rotB / kT);
      }
      Qpf /= (double)sumSw;
      QpfZ = Qpf * exp(zpe / kT);
      ouf << (10000. / T) << ',' << T << ',' << Qpf << ',' << QpfZ
       << endl;
    }
  }
  delete ofmf;
  ouf.close();
  return NORM;
}

void basisAssocLegendre::clear_ab(int m) {           // --- clear_ab ---
  int moPE = potE.maxOrder(), irow, jcol;
  for (irow = 0; irow <= moPE; irow++) {
    for (jcol = 0; jcol <= maxJdeg - m; jcol++) { ab[jcol][irow] = 0.; }
  }
}
                                                        // --- kappa ---
vector<double> basisAssocLegendre::kappa(int m, int kc) {
  int iord, moPE = potE.maxOrder();
  vector<double> res(maxJdeg - m + 1, 0.), cfPE = potE.coefVect();
  recurExpnCoef rcof(m, maxJdeg, kc);
  rcof.addTo(res, cfPE[0]);
  for (iord = 1; iord <= moPE; iord++)
   { rcof.incPow(); rcof.addTo(res, cfPE[iord]); }
  return res;
}

void basisAssocLegendre::set_ab_L(int m) {           // --- set_ab_L ---
  int moPE = potE.maxOrder(), mSiz = maxJdeg - m + 1;
  int j, jcol, irow, irowmax;
  vector<double> kpp;

  clear_ab(m);
  for (j = m; j <= maxJdeg; j++) {
    jcol = j - m;
    irowmax = jcol + moPE;
    if (irowmax > (mSiz - 1)) { irowmax = mSiz - 1; }
    kpp = kappa(m, j);
    for (irow = jcol; irow <= irowmax; irow++) {
      ab[jcol][irow - jcol] = kpp[irow];
    }
    ab[jcol][0] += j * (j + 1);
  }
}

int basisAssocLegendre::solveForOneM(int m) {    // --- solveForOneM ---
  int mSiz = maxJdeg - m + 1, nOffDg = potE.maxOrder(), info;
  mJcur = -1; set_ab_L(m);
  info = lapack_dsbev('V', 'L', mSiz, nOffDg, (double **)ab,
                      MXRWAB, w, (double **)z, MXNBAS, work);
  if (info < 0) {
    cout << "DSBEV: - invalid input.\n"; return ERRE;
  } else if (info > 0) {
    cout << "DSBEV: - failed to converge.\n"; return ERRE;
  }
  mJcur = m;
  return NORM;
}

                                                 // --- writeVectors ---
int basisAssocLegendre::writeVectors(string &basfln) {
  if (mJcur < 0) { return ERRE; }
  int nPages = (outJmax - mJcur) / 250 + 1, jp, jbegin, jlast, ij, ib;
  string outfln;
  ofstream ouf;
  outFormat *ofmt;
  for (jp = 0; jp < nPages; jp++) {
    jbegin = jp * 250 + mJcur; jlast = jp * 250 + 249;
    if (jlast > outJmax) { jlast = outJmax; }
    outfln = basfln + "_eigen_vectors_mJ" + format(mJcur, 4, '0') + "_p"
           + format(jp, 2, '0') + ".csv";
    ouf.open(outfln.c_str());
    if (!ouf) { return ERRE; }
    ofmt = new outFormat(&ouf);
    ouf << "eigen_values"; ouf.precision(8);
    for (ij = jbegin; ij <= jlast; ij++) { ouf << "," << w[ij - mJcur]; }
    ouf << "\neigen_vectors" ;
    for (ij = jbegin; ij <= jlast; ij++) { ouf << ",vec" << ij; }
    ouf << endl; ouf.precision(4);
    for (ib = mJcur; ib <= maxJdeg; ib++) {
      ouf << "c" << ib;
      for (ij = jbegin; ij <= jlast; ij++)
        { ouf << "," << z[ij - mJcur][ib - mJcur]; }
      ouf << endl;
    }
    delete ofmt;
    ouf.close();
  }
  return NORM;
}

                                               // --- writeFunctions ---
int basisAssocLegendre::writeFunctions(string &basfln) {
  if (mJcur < 0) { return ERRE; }
  int nPages = (outJmax - mJcur) / 250 + 1, jp, jbegin, jlast, ij, ic;
  int nZdiv = outJmax * 5, iz;
  double zv, f;
  vector<double> nPnm;
  string outfln;
  vector<ofstream *> ouf(nPages);
  ofstream *ouf1;
  vector<outFormat *> ofmt(nPages);
  outFormat *ofmt1;
  ouf.clear(); ofmt.clear();
  for (jp = 0; jp < nPages; jp++) {
    jbegin = jp * 250 + mJcur; jlast = jp * 250 + 249;
    if (jlast > outJmax) { jlast = outJmax; }
    outfln = basfln + "_eigen_funcs_mJ" + format(mJcur, 4, '0') + "_p"
           + format(jp, 2, '0') + ".csv";
    ouf1 = new ofstream(outfln.c_str());
    if (!(*ouf1)) { return ERRE; }
    ouf.push_back(ouf1);
    ofmt1 = new outFormat(ouf[jp]);
    ofmt.push_back(ofmt1);
    *ouf[jp] << "z";
    for (ij = jbegin; ij <= jlast; ij++) { *ouf[jp] << ",f" << ij; }
    *ouf[jp] << endl;
  }
  for (iz = 0; iz <= nZdiv; iz++) {
    zv = iz * 2. / nZdiv - 1.;
    nPnm = nrmPnm(mJcur, maxJdeg, zv);
    for (jp = 0; jp < nPages; jp++) {
      jbegin = jp * 250 + mJcur; jlast = jp * 250 + 249;
      if (jlast > outJmax) { jlast = outJmax; }
      ouf[jp]->precision(6); *ouf[jp] << zv; ouf[jp]->precision(4);
      for (ij = jbegin; ij <= jlast; ij++) {
        f = 0.;
        for (ic = mJcur; ic <= maxJdeg; ic++)
         { f += z[ij - mJcur][ic - mJcur] * nPnm[ic - mJcur]; }
        *ouf[jp] << "," << f;
      }
      *ouf[jp] << endl;
    }
  }
  for (jp = 0; jp < nPages; jp++) {
    delete ofmt[jp]; ouf[jp]->close(); delete ouf[jp];
  }
  return NORM;
}

                                                        // --- errms ---
int basisAssocLegendre::errms(errcode ec, string ms) {
  string s;
  switch(ec) {
    case FAILopen: s = "- Failed to open [" + ms + "]."; break;
    case LRGQerr: s = "- No partfunc calc can be performed."; break;
    case LMTQrng: s = "warning: Limited T-range for Q-calc."; break;
  }
  cout << s << endl;
  return ERRE;
}

