/* ======================================================================
     "libbx1PS.cc"
       - basis expansion library for power series potential and
          Hermite-Gaussian basis
             Copyright (c) 2004-2005  by A. Miyoshi, Univ. Tokyo
                         created: Sept.  8, 2005
                          edited: Sept. 17, 2005   Rev. 0.10
                     last edited: Oct.  15, 2005   Rev. 0.11
 ------------------------------------------------------------------------
 - Rev. 0.11 : added member nSpinW (nucSpinStat class object)
                to basisHermiteGauss class
 ====================================================================== */

#include "libbx1PS.h"

/* ======================================================================
     extended double variable class
 ====================================================================== */

                                                 // --- Constructors ---
extDouble::extDouble() { clear(); }
extDouble::extDouble(int s, double lv) { set(s, lv); }
extDouble::extDouble(double v) { set(v); }

                                                // --- intitializers ---
void extDouble::clear() { sign = 1; lnval = 0.; }
void extDouble::set(int s, double lv) { sign = s; lnval = lv; }
void extDouble::set(double v) {
  if (v > 0.) { sign = 1; lnval = log(v); }
  else if (v < 0.) { sign = -1; lnval = log(-v); }
  else { sign = 0; lnval = 0.; }
}

double extDouble::val() { return sign * exp(lnval); }     // --- val ---

extDouble extDouble::operator*(double v) {           // --- * double ---
  extDouble rs;
  int vsign;
  double vlnval;
  if (v > 0.) { vsign = 1; vlnval = log(v); }
  else if (v < 0.) { vsign = -1; vlnval = log(-v); }
  else { vsign = 0; vlnval = 0.; }
  rs.sign = sign * vsign;
  if (rs.sign == 0) { rs.lnval = 0.; }
  else { rs.lnval = lnval + vlnval; }
  return rs;
}

extDouble extDouble::operator*(extDouble lev) {   // --- * extDouble ---
  extDouble rs;
  rs.sign = sign * lev.sign;
  if (rs.sign == 0) { rs.lnval = 0.; }
  else { rs.lnval = lnval + lev.lnval; }
  return rs;
}

extDouble extDouble::operator+(extDouble lev) {   // --- + extDouble ---
  extDouble rs;
  if (sign == 0) {
    rs.sign = lev.sign; rs.lnval = lev.lnval;
  } else if (lev.sign == 0) {
    rs.sign = sign; rs.lnval = lnval;
  } else if (sign == lev.sign) {
    rs.sign = sign;
    rs.lnval = lnval + log(1. + exp(lev.lnval - lnval));
  } else {
    if (lnval > lev.lnval) {
      rs.sign = sign;
      rs.lnval = lnval + log(1. - exp(lev.lnval - lnval));
    } else if (lnval < lev.lnval) {
      rs.sign = lev.sign;
      rs.lnval = lev.lnval + log(1. - exp(lnval - lev.lnval));
    } else {
      rs.sign = 0; rs.lnval = 0.;
    }
  }
  return rs;
}

/* ======================================================================
     power series potential and related classes
 ====================================================================== */

powerSeries::powerSeries() { clear(); maxOrder = 0; }

void powerSeries::setMaxOrder() {
  iterator p;
  maxOrder = 0;
  for (p = begin(); p < end(); p++) {
    if (p->order > maxOrder) { maxOrder = p->order; }
  }
}

/* ======================================================================
     Hermite-Gaussian basis set and related classes
 ====================================================================== */

/* ----- integCoefHGPS and related ---------------------------------- */

double termIdxHGPS::value(int k) {
  int nidx = size(), id;
  double v = 1.;
  for (id = 0; id < nidx; id++) { v *= (k + (*this)[id]); }
  return sqrt(v);
}

double coefTermsHGPS::value(int k) {
  int ntms = size(), it;
  double v = 0.;
  for (it = 0; it < ntms; it++) { v += (*this)[it].value(k); }
  return v;
}

integCoefHGPS::integCoefHGPS() { clear(); maxOrder = 0; }

integCoefHGPS::integCoefHGPS(int mord) { setCoefIndices(mord); }

void integCoefHGPS::setCoefIndices(int mord) {
  int iord = 0, jord = 0, liord, ljord, ntms, addidx, it;
  termIdxHGPS term;
  coefTermsHGPS coef, lcoef;
  map<int, coefTermsHGPS> coefmap;
  clear();
  maxOrder = mord;
   // set iord = 0
  term.clear(); coef.clear(); coef.push_back(term);
  coefmap.clear(); coefmap.insert(make_pair(jord, coef));
  push_back(coefmap);
   // set iord > 0
  for (iord = 1; iord <= maxOrder; iord++) {
    liord = iord - 1;
    coefmap.clear();
    for (jord = -iord; jord <= iord; jord += 2) {
      coef.clear();
      for (ljord = (jord - 1); ljord <= (jord + 1); ljord += 2) {
        if ((ljord >= -liord) && (ljord <= liord)) {
          if (ljord < jord) { addidx = jord; } else { addidx = ljord; }
          lcoef = ((*this)[liord])[ljord];
          ntms = lcoef.size();
          for (it = 0; it < ntms; it++) {
            lcoef[it].push_back(addidx);
            coef.push_back(lcoef[it]);
          }
        }
      }
      coefmap.insert(make_pair(jord, coef));
    }
    push_back(coefmap);
  }
}

/* ----- basisHermiteGauss ------------------------------------------ */

                                                 // --- Constructors ---
basisHermiteGauss::basisHermiteGauss() { init(); }

basisHermiteGauss::basisHermiteGauss(int nbas, int mord, double B,
 double alp, nucSpinStat &nsw)
 { init(); setParam(nbas, mord, B, alp, nsw); }

                                                // --- intitializers ---
void basisHermiteGauss::init() {
  int i;
  nBasis = 0; rotB = 0.; alpha = 0.; nSpinW.clear(); cInteg.clear();
  cH0.set(1, -log(M_PI) / 4.);
  cH1.set(1, log(2.) / 2. - log(M_PI) / 4.);
  for (i = 2; i < MXNBAS; i++) {
    c[i][0].set(1, (log(2.) - log((double)i)) / 2.);
    c[i][1].set(-1, (log((double)(i - 1)) - log((double)i)) / 2.);
  }
}

void basisHermiteGauss::setParam(int nbas, int mord, double B,
 double alp, nucSpinStat &nsw) {
  nBasis = nbas; rotB = B; alpha = alp; nSpinW = nsw;
  cInteg.setCoefIndices(mord);
}

int basisHermiteGauss::numBasis() { return nBasis; }   // --- nBasis ---

int basisHermiteGauss::setNumBasis(int nbas) {     // --- set nBasis ---
  if (nbas > MXNBAS) { return ERRE; }
  nBasis = nbas;
  return NORM;
}

double basisHermiteGauss::HGalpha() { return alpha; }   // --- alpha ---

int basisHermiteGauss::solve(powerSeries &potE) {       // --- solve ---
  int mxOrd = potE.maxOrder, nPotp = potE.size();
  double alp22 = 2. * alpha * alpha, aosq2 = alpha / sqrt(2.);
  int nOffDg, kShft, j, k, l, lmx, ip, n, nchk, info;
  double apot, D, U;

  nOffDg = (mxOrd > 2) ? mxOrd : 2;
  for (k = 0; k < nBasis; k++) {
    lmx = k + nOffDg;
    if (lmx > (nBasis - 1)) { lmx = (nBasis - 1); }
    for (l = k; l <= lmx; l++) {
      D = 0.; U = 0.; j = l - k;
      if (j == 0) { D = -(2. * k + 1.) / alp22; }
      else if (j == 2) { D = sqrt((l - 1.) * l) / alp22; }
      if (j <= mxOrd) {
        for (ip = 0; ip < nPotp; ip++) {
          n = potE[ip].order; apot = potE[ip].coef;
          nchk = n - j;
          if ((nchk >= 0) && ((nchk % 2) == 0))
           { U += apot * pow(aosq2, n) * (cInteg[n])[j].value(k); }
        }
      }
      kShft = nOffDg - l + k;
      ab[l][kShft] = U - rotB * D;
    }
  }
  info = lapack_dsbev('V', 'U', nBasis, nOffDg, (double **)ab,
                      MXRWAB, w, (double **)z, MXNBAS, work);
  if (info < 0) { cout << "DSBEV: - invalid input.\n"; return ERRE; }
  else if (info > 0)
   { cout << "DSBEV: - failed to converge.\n"; return ERRE; }
  return NORM;
}

                                             // --- get eigen values ---
double *basisHermiteGauss::getEigenValues() { return w; }

          // --- get eigen vector coefficients of basis function ibf ---
double *basisHermiteGauss::getEgnVecCoefs(int ibf) {
  int iq;
  for (iq = 0; iq < nBasis; iq++) { work[iq] = z[iq][ibf]; }
  return work;
}

                                                 // --- getEigenFunc ---
double *basisHermiteGauss::getEigenFunc(double y, int mqn) {
  int ic, ir;
  double wvf;
  setBasFunc(y, nBasis);
  for (ir = 0; ir <= mqn; ir++) {
    wvf = 0.;
    for (ic = 0; ic < nBasis; ic++) { wvf += z[ir][ic] * b[ic]; }
    f[ir] = wvf;
  }
  return f;
}

                                                   // --- setBasFunc ---
void basisHermiteGauss::setBasFunc(double y, int nbas) {
  int i;
  extDouble ye(y), Hn0, Hn1, HnC;
  Hn0 = cH0; Hn0.lnval -= (log(alpha) + y * y) / 2.;
  Hn1 = cH1 * ye; Hn1.lnval -= (log(alpha) + y * y) / 2.;
  b[0] = Hn0.val();
  b[1] = Hn1.val();
  for (i = 2; i < nbas; i++) {
    HnC = ye * c[i][0] * Hn1 + c[i][1] * Hn0;
    b[i] = HnC.val();
    Hn0 = Hn1; Hn1 = HnC;
  }
}

