/* ======================================================================
     "libnumLls_dsysv.cc"
       - Linear least squares library (test version using DSYSV
          instead of DPOSV)
             Copyright (c) 2005-2007  by A. Miyoshi, Univ. Tokyo
                         created: Jan.   2, 2007
                from Oct. 11, 2005 (Rev. 0.11) version of libnumLls
                         created: Jan.   2, 2007
                     last edited: Jan.   2, 2007
 ====================================================================== */

#include "libnumLls_dsysv.h"

/* ======================================================================
     basic Linear Least Squares Class
 ====================================================================== */

                                                 // --- Constructors ---
basicLinLsSq::basicLinLsSq() { clear(); }

basicLinLsSq::basicLinLsSq(int npar) { init(npar); }

                                                 // --- Initializers ---
void basicLinLsSq::clear()
 { nPars = 0; nMeas = 0; sumXX.clear(); sumYX.clear(); }

int basicLinLsSq::init(int npar) {
  int ip;
  clear();
  if (npar > MXNPAR) { errmsg(INIT, LRG_nPars); return ERRE; }
  else if (npar < 1) { errmsg(INIT, SML_nPars); return ERRE; }
  nPars = npar;
  for (ip = 0; ip < nPars; ip++)
   { sumYX.push_back(0.); sumXX.push_back(sumYX); }
  sumYY = 0.;
  return NORM;
}

                                                // --- Data addition ---
int basicLinLsSq::addDatumW(double y, vector<double> &x, double w) {
  int ip, iq;
  if (nPars == 0) { errmsg(INIT, NOT_init); return ERRE; }
  if (x.size() < nPars) { errmsg(DADD, SMLN_x); return ERRE; }
  for (ip = 0; ip < nPars; ip++) {
    sumYX[ip] += w * x[ip] * y;
    for (iq = 0; iq <= ip; iq++) { sumXX[ip][iq] += w * x[ip] * x[iq]; }
  }
  sumYY += w * y * y;
  nMeas++;
  return NORM;
}

int basicLinLsSq::addDatum(double y, vector<double> &x) {
  double w = 1.;
  return addDatumW(y, x, w);
}

int basicLinLsSq::putDataW(int n, vector<double> &y,
 vector<vector<double> > &x, vector<double> &w) {
  int id;
  if ((y.size() < n) || (x.size() < n) || (w.size() < n))
   { errmsg(DADD, SMLN_meas); return ERRE; }
  for (id = 0; id < n; id++) {
    if (addDatumW(y[id], x[id], w[id]) == ERRE) { return ERRE; }
  }
  return NORM;
}

int basicLinLsSq::putData(int n, vector<double> &y,
 vector<vector<double> > &x) {
  int id;
  if ((y.size() < n) || (x.size() < n))
   { errmsg(DADD, SMLN_meas); return ERRE; }
  for (id = 0; id < n; id++) {
    if (addDatum(y[id], x[id]) == ERRE) { return ERRE; }
  }
  return NORM;
}

                                               // --- Solve basicLLS ---
int basicLinLsSq::solve_blls(vector<double> &mlp, vector<double> &sdp,
 vector<vector<double> > &pcc) {
  int ir, ic, info;
  double Vy, sdy = 0.;
  vector<double> wcc;

  if (nPars == 0) { errmsg(INIT, NOT_init); return ERRE; }
  if (nMeas < nPars) { errmsg(SOLV, meas_SMLTp); return ERRE; }
  for (ir = 0; ir < nPars; ir++) {
    b[0][ir] = sumYX[ir];
    for (ic = 0; ic <= ir; ic++) { a[ic][ir] = sumXX[ir][ic]; }
  }
//
//  info = lapack_dsysv('L', nPars, 1, (double **)a, MXNPAR, ipiv,
//   (double **)b, MXNPAR, work, -1);
//  cout << "work[0]:" << work[0] << endl;
//
  info = lapack_dsysv('L', nPars, 1, (double **)a, MXNPAR, ipiv,
   (double **)b, MXNPAR, work, MXLWRK);
  if (info < 0) { errmsg(SOLV, POSVpar); return ERRE; }
  else if (info > 0) { errmsg(SOLV, POSVfail); return ERRE; }
  info = lapack_dsytri('L', nPars, (double **)a, MXNPAR, ipiv, work);
  if (info < 0) { errmsg(SOLV, POTRIpar); return ERRE; }
  else if (info > 0) { errmsg(SOLV, POTRIfail); return ERRE; }

  Vy = sumYY;
  mlp.clear(); sdp.clear(); pcc.clear();
  for (ir = 0; ir < nPars; ir++) {
    mlp.push_back(b[0][ir]);
    Vy += mlp[ir] * mlp[ir] * sumXX[ir][ir] - 2. * mlp[ir] * sumYX[ir];
    sdp.push_back(sqrt(a[ir][ir]));
    wcc.clear();
    for (ic = 0; ic < ir; ic++) {
      wcc.push_back(a[ic][ir] / sqrt(a[ir][ir] * a[ic][ic]));
      Vy += 2. * mlp[ir] * mlp[ic] * sumXX[ir][ic];
    }
    pcc.push_back(wcc);
  }
  if (nMeas > nPars) { sdy = sqrt(Vy / (nMeas - nPars)); }
  for (ir = 0; ir < nPars; ir++) { sdp[ir] *= sdy; }
  return NORM;
}

                                                // --- Error message ---
void basicLinLsSq::errmsg(erract era, errcode erc) {
  string msg = "basicLinLsSq-error: ";
  switch(era) {
    case INIT: msg += "(initialization) "; break;
    case DADD: msg += "(data addition) "; break;
    case SOLV: msg += "(solver) "; break;
  }
  switch(erc) {
    case NOT_init: msg += "object not initialized"; break;
    case LRG_nPars: msg += "nPars larger than MXNPAR"; break;
    case SML_nPars: msg += "nPars smaller than 1"; break;
    case SMLN_x: msg += "x-vec smaller than # of parame"; break;
    case SMLN_meas: msg += "meas-vec smaller than # of meas"; break;
    case meas_SMLTp: msg += "# of meas smaller than # of param"; break;
    case POSVpar: msg += "invalid input to DPOSV"; break;
    case POSVfail: msg += "DPOSV numerical error"; break;
    case POTRIpar: msg += "invalid input to DPOTRI"; break;
    case POTRIfail: msg += "DPOTRI numerical error"; break;
  }
  cout << msg << endl;
}

/* ======================================================================
     Multiple Dependent-Variable Linear Least Squares
      and Related Classes
 ====================================================================== */

// --- function unit ---------------------------------------------------

funcUnit::funcUnit() { clear(); }                // --- constructors ---

funcUnit::funcUnit(string nam, int ord, double cf)
 { init(nam, ord, cf); }

void funcUnit::clear()                           // --- initializers ---
 { fname = ""; order = 0; coef = 0.; }

int funcUnit::init(string nam, int ord, double cf) {
  string snam = strAllToLower(nam);
  fname = snam; order = ord; coef = cf;
  if ((snam == "null") || (fabs(cf) < VAL0THR)) {
    fname = "null"; order = 0; coef = 0.;
  } else if ((snam == "const") || (snam == "ln")) {
    order = 0;
  } else if ((snam == "pow") || (snam == "cos")) {
    if (ord == 0) { fname = "const"; }
  } else if (snam == "sin") {
    if (ord == 0) { fname = "null"; coef = 0.; }
  } else {
    fname = ""; order = 0; coef = 0.; return ERRE;
  }
  return NORM;
}

int funcUnit::isProperlySet() {             // --- is properly set ? ---
  if (fname == "") { return false; }
  return true;
}

string funcUnit::termStr() {         // --- term representing string ---
  string sord = format(order), scoef = format(coef), sfunc;
  if (fname == "null") { return ""; }
  else if (fname == "const") { return scoef; }
  if (scoef == "1") { scoef = ""; } else { scoef += "*"; }
  if (fname == "ln") { sfunc = "ln(x)"; }
  else if (fname == "pow") {
    if (order == 1) { sfunc = "x"; }
    else { sfunc = "x^" + sord; }
  }
  else if ((fname == "sin") || (fname == "cos"))
   { sfunc = fname + "(" + sord + "*x)"; }
  else { return "?"; }
  return scoef + sfunc;
}

double funcUnit::val(double x) {          // --- function value at x ---
  if (fname == "null") { return 0.; }
  else if (fname == "const") { return coef; }
  else if (fname == "pow") { return coef * pow(x, order); }
  else if (fname == "ln") { return coef * log(x); }
  else if (fname == "sin") { return coef * sin(order * x); }
  else if (fname == "cos") { return coef * cos(order * x); }
  return 0.;
}

funcUnit funcUnit::differnFunc() {     // --- differentiate function ---
  int dford;
  double dfcf;
  string dfnam;
  funcUnit derf;
  if (fname == "ln") {
    dfnam = "pow"; dford = -1; dfcf = coef;
  } else if (fname == "pow") {
    dfnam = fname; dford = order - 1; dfcf = coef * order;
  } else if (fname == "sin") {
    dfnam = "cos"; dford = order; dfcf = coef * order;
  } else if (fname == "cos") {
    dfnam = "sin"; dford = order; dfcf = -coef * order;
  } else {
    dfnam = "null"; dford = 0; dfcf = 0.;
  }
  derf.init(dfnam, dford, dfcf);
  return derf;
}

// --- list of funcUnit's = a sigle variable function ------------------

                                                   // --- add a term ---
int listFnUnits::add(string nam, int ord, double cf) {
  funcUnit trm(nam, ord, cf);
  if (!trm.isProperlySet()) { return ERRE; }
  push_back(trm);
  return NORM;
}

int listFnUnits::addNoOrd(string nam, double cf) {
  funcUnit trm(nam, 0, cf);
  if (!trm.isProperlySet()) { return ERRE; }
  push_back(trm);
  return NORM;
}
                                             // --- add from strings ---
int listFnUnits::addFrom(vector<string> &vs) {
  int numStkns = vs.size();
  if (numStkns == 0) {
    return ERRE;
  } else if (numStkns == 1) {
    return addNoOrd(vs[0]);
  } else if (numStkns == 2) {
    if ((vs[0] == "null") || (vs[0] == "const") || (vs[0] == "ln"))
     { return addNoOrd(vs[0], atof(vs[1].c_str())); }
    else { return add(vs[0], atoi(vs[1].c_str())); }
  }
  return add(vs[0], atoi(vs[1].c_str()), atof(vs[2].c_str()));
}

listFnUnits listFnUnits::differnFunc() {        // --- differentiate ---
  listFnUnits derf;
  listFnUnits::iterator p;
  derf.clear();
  for (p = begin(); p != end(); p++) { derf.push_back(p->differnFunc()); }
  return derf;
}

// --- multiple dependent-variable Linear Least Squares ----------------

multiVarLLS::multiVarLLS() { clear(); }          // --- constructors ---

multiVarLLS::multiVarLLS(int ndv) { init(ndv); }

void multiVarLLS::clear() {                      // --- initializers ---
  basicLinLsSq::clear();
  optFuncs.clear(); optCoef.clear();
  fixFuncs.clear(); fixCoef.clear();
}

int multiVarLLS::init(int ndv) {
  int it;
  listFnUnits trm;
  clear();
  if (ndv < 1) { return loc_err(INVinitPAR); }
  trm.clear();
  for (it = 0; it < ndv; it++)
   { optFuncs.push_back(trm); fixFuncs.push_back(trm); }
  return NORM;
}

int multiVarLLS::numDpndVars()  // --- number of dependent variables ---
 { return optFuncs.size(); }
                                             // --- set one opt-Func ---
int multiVarLLS::set1OptFnc(int fnbr, listFnUnits &lfu) {
  if (numDpndVars() == 0) { return loc_err(NOTinitOBJ); }
  if ((fnbr < 0) || (fnbr >= optFuncs.size()))
   { return loc_err(INVfuncNBR); }
  if (lfu.size() == 0) { return loc_err(INVfuncLST); }
  optFuncs[fnbr] = lfu;
  return NORM;
}
                                       // --- add each opt-Func term ---
int multiVarLLS::add1OptFncTrm(int fnbr, string nam, int ord,
 double cf) {
  if (numDpndVars() == 0) { return loc_err(NOTinitOBJ); }
  if ((fnbr < 0) || (fnbr >= optFuncs.size()))
   { return loc_err(INVfuncNBR); }
  return optFuncs[fnbr].add(nam, ord, cf);
}

int multiVarLLS::add1OptFncTrmNoOrd(int fnbr, string nam, double cf)
 { return add1OptFncTrm(fnbr, nam, 0, cf); }

int multiVarLLS::add1OptFncTrmFrom(int fnbr, vector<string> &vs) {
  if (numDpndVars() == 0) { return loc_err(NOTinitOBJ); }
  if ((fnbr < 0) || (fnbr >= optFuncs.size()))
   { return loc_err(INVfuncNBR); }
  return optFuncs[fnbr].addFrom(vs);
}
                                     // --- set one fix-Func & coefs ---
int multiVarLLS::set1FixFncCoef(int fnbr, listFnUnits &lfu,
 vector<double> &ffc) {
  if (numDpndVars() == 0) { return loc_err(NOTinitOBJ); }
  if ((fnbr < 0) || (fnbr >= fixFuncs.size()))
   { return loc_err(INVfuncNBR); }
  if (lfu.size() == 0) { return NORM; }
  if (ffc.size() < lfu.size()) { return loc_err(INSnumCOEF); }
  fixFuncs[fnbr] = lfu;
  fixCoef = ffc;
  return NORM;
}
                                       // --- add each fix-Func term ---
int multiVarLLS::add1FixFncTrm(int fnbr, string nam, int ord,
 double cf) {
  if (numDpndVars() == 0) { return loc_err(NOTinitOBJ); }
  if ((fnbr < 0) || (fnbr >= fixFuncs.size()))
   { return loc_err(INVfuncNBR); }
  return fixFuncs[fnbr].add(nam, ord, cf);
}

int multiVarLLS::add1FixFncTrmNoOrd(int fnbr, string nam, double cf)
 { return add1FixFncTrm(fnbr, nam, 0, cf); }

int multiVarLLS::add1FixFncTrmFrom(int fnbr, vector<string> &vs) {
  if (numDpndVars() == 0) { return loc_err(NOTinitOBJ); }
  if ((fnbr < 0) || (fnbr >= fixFuncs.size()))
   { return loc_err(INVfuncNBR); }
  return fixFuncs[fnbr].addFrom(vs);
}

void multiVarLLS::addFixCoef(double fc) {        // --- add fix-coef ---
  fixCoef.push_back(fc);
}
                                      // --- set derivative function ---
int multiVarLLS::setDerivFunc(int fnbr, int gnbr, int deg) {
  if (numDpndVars() == 0) { return loc_err(NOTinitOBJ); }
  if (    (fnbr < 0) || (fnbr >= numDpndVars())
       || (gnbr < 0) || (gnbr >= numDpndVars()) )
   { return loc_err(INVfuncNBR); }
  if (optFuncs[gnbr].size() == 0) { return loc_err(NOTsetGFNC); }
  if (deg <= 0) { return loc_err(INVdegDIFF); }
  listFnUnits dopf = optFuncs[gnbr], dfxf = fixFuncs[gnbr];
  int id;
  for (id = 0; id < deg; id++)
   { dopf = dopf.differnFunc(); dfxf = dfxf.differnFunc(); }
  optFuncs[fnbr] = dopf; fixFuncs[fnbr] = dfxf;
  return NORM;
}

int multiVarLLS::numTermsOpt() {         // --- nbr of optFunc terms ---
  if (numDpndVars() == 0) { return 0; }
  return optFuncs[0].size();
}

int multiVarLLS::numTermsFix() {         // --- nbr of fixFunc terms ---
  if (numDpndVars() == 0) { return 0; }
  return fixFuncs[0].size();
}

                                 // --- function representing string ---
string multiVarLLS::funcStr(int fnbr) {
  int ntOp = numTermsOpt(), ntFx = numTermsFix(), jop, jfx;
  string fs = "", ts;
  if ((fnbr < 0) || (fnbr >= numDpndVars())) { return ""; }
  for (jop = 0; jop < ntOp; jop++) {
    ts = optFuncs[fnbr][jop].termStr();
    if (ts != "") {
      if (fs == "") { fs += " = "; } else { fs += " + "; }
      fs += "a" + format(jop);
      if (ts != "1") { fs += " * " + ts; }
    }
  }
  for (jfx = 0; jfx < ntFx; jfx++) {
    ts = fixFuncs[fnbr][jfx].termStr();
    if (ts != "") {
      if (fs == "") { fs += " = "; } else { fs += " + "; }
      fs += "c" + format(jfx);
      if (ts != "1") { fs += " * " + ts; }
    }
  }
  if (fs == "") { fs += " = 0"; }
  return fs;
}
                            // --- all function representing strings ---
vector<string> multiVarLLS::funcsStrAll() {
  int idv, ndv = numDpndVars();
  vector<string> vs;
  vs.clear();
  for (idv = 0; idv < ndv; idv++) { vs.push_back(funcStr(idv)); }
  return vs;
}

vector<double> multiVarLLS::getFixCoef()   // --- fixed coefficients ---
 { return fixCoef; }

int multiVarLLS::isFuncSet() {              // --- is function set ? ---
  int numOpFnc = optFuncs.size(), numFxFnc = fixFuncs.size(), ntOp, iopf;
  int ntFx, ifxf;
  if (numOpFnc == 0) { return false; }
  if (numFxFnc < numOpFnc) { return false; }
  if ((ntOp = optFuncs[0].size()) == 0) { return false; }
  for (iopf = 1; iopf < numOpFnc; iopf++) {
    if (optFuncs[iopf].size() < ntOp) { return false; }
  }
  ntFx = fixFuncs[0].size();
  for (ifxf = 1; ifxf < numFxFnc; ifxf++) {
    if (fixFuncs[ifxf].size() < ntFx) { return false; }
  }
  if (fixCoef.size() < ntFx) { return false; }
  return true;
}
                                            // --- add a datum point ---
int multiVarLLS::addPoint(int fnbr, double x, double y, double w) {
  int ntOp = numTermsOpt(), ntFx = numTermsFix(), jop, jfx;
  double ysub;
  vector<double> vx;
  if (!isFuncSet()) { return loc_err(NOTsetFNCS); }
  if ((fnbr < 0) || (fnbr >= numDpndVars()))
   { return loc_err(INVfuncNBR); }
  if (nPars < ntOp) {
    if (basicLinLsSq::init(ntOp) == ERRE) { return ERRE; }
  }
  ysub = y;
  for (jfx = 0; jfx < ntFx; jfx++)
   { ysub -= fixCoef[jfx] * fixFuncs[fnbr][jfx].val(x); }
  vx.clear();
  for (jop = 0; jop < ntOp; jop++)
   { vx.push_back(optFuncs[fnbr][jop].val(x)); }
  return addDatumW(ysub, vx, w);
}
                                              // --- set data points ---
int multiVarLLS::setPoints(int fnbr, dataPntsXYW &dps) {
  int np = dps.size(), ip;
  for (ip = 0; ip < np; ip++) {
    if (addPoint(fnbr, dps[ip].x, dps[ip].y, dps[ip].w) == ERRE)
     { return ERRE; }
  }
  return NORM;
}
                                                    // --- Solve LLS ---
int multiVarLLS::solve(vector<double> &mlp, vector<double> &sdp,
 vector<vector<double> > &pcc) {
  int retc;
  retc = solve_blls(mlp, sdp, pcc);
  if (retc == NORM) { optCoef = mlp; }
  return retc;
}

int multiVarLLS::wasSolved() {                   // --- was solved ? ---
  if (optCoef.size() > 0) { return true; } else { return false; }
}
                                            // --- is power series ? ---
int multiVarLLS::isPowerSer(int fnbr, powFourSer &ps) {
  int ntOp = numTermsOpt(), ntFx = numTermsFix(), jop, jfx;
  string nam;
  powFourUnit pu;
  ps.clear();
  if (!wasSolved()) { return false; }
  if ((fnbr < 0) || (fnbr >= numDpndVars()))
   { loc_err(INVfuncNBR); return false; }
  for (jop = 0; jop < ntOp; jop++) {
    nam = optFuncs[fnbr][jop].fname; pu.order = optFuncs[fnbr][jop].order;
    pu.coef = optFuncs[fnbr][jop].coef; pu.param = optCoef[jop];
    if ((nam == "const") || (nam == "pow")) { ps.push_back(pu); }
    else if (nam != "null") { return false; }
  }
  for (jfx = 0; jfx < ntFx; jfx++) {
    nam = fixFuncs[fnbr][jfx].fname; pu.order = fixFuncs[fnbr][jfx].order;
    pu.coef = fixFuncs[fnbr][jfx].coef; pu.param = fixCoef[jfx];
    if ((nam == "const") || (nam == "pow")) { ps.push_back(pu); }
    else if (nam != "null") { return false; }
  }
  return true;
}
                                          // --- is Fourier series ? ---
int multiVarLLS::isFourierSer(int fnbr, powFourSer &fs) {
  int ntOp = numTermsOpt(), ntFx = numTermsFix(), jop, jfx;
  string nam;
  powFourUnit fu;
  fs.clear();
  if (!wasSolved()) { return false; }
  if ((fnbr < 0) || (fnbr >= numDpndVars()))
   { loc_err(INVfuncNBR); return false; }
  for (jop = 0; jop < ntOp; jop++) {
    nam = optFuncs[fnbr][jop].fname; fu.order = optFuncs[fnbr][jop].order;
    fu.coef = optFuncs[fnbr][jop].coef; fu.param = optCoef[jop];
    if ((nam == "const") || (nam == "cos")) {
      fu.order = abs(fu.order); fs.push_back(fu);
    } else if (nam == "sin") {
      if (fu.order < 0) { fu.coef = -fu.coef; }
      fu.order = -abs(fu.order); fs.push_back(fu);
    } else if (nam != "null") { return false; }
  }
  for (jfx = 0; jfx < ntFx; jfx++) {
    nam = fixFuncs[fnbr][jfx].fname; fu.order = fixFuncs[fnbr][jfx].order;
    fu.coef = fixFuncs[fnbr][jfx].coef; fu.param = fixCoef[jfx];
    if ((nam == "const") || (nam == "cos")) {
      fu.order = abs(fu.order); fs.push_back(fu);
    } else if (nam == "sin") {
      if (fu.order < 0) { fu.coef = -fu.coef; }
      fu.order = -abs(fu.order); fs.push_back(fu);
    } else if (nam != "null") { return false; }
  }
  return true;
}
                                          // --- local error message ---
int multiVarLLS::loc_err(loc_errcod lerc) {
  string msg = "multiVarLLS-error: ";
  switch(lerc) {
    case INVinitPAR: msg += "invalid initialization parameter"; break;
    case NOTinitOBJ: msg += "object not initialized"; break;
    case INVfuncNBR: msg += "invalid function number"; break;
    case INVfuncLST: msg += "invalid function list"; break;
    case INSnumCOEF: msg += "insufficient number of coefficients"; break;
    case NOTsetGFNC: msg += "generating function not set"; break;
    case INVdegDIFF: msg += "invalid degree of defferentiation"; break;
    case NOTsetFNCS: msg += "function not set"; break;
  }
  cout << msg << endl;
  return ERRE;
}

